# =====================================================================================================
# Basic tool to dump content information and/or samples out of a DTPK file, 
# as well splicing in compatible PCM WAV files or YADPCM chunks.
# This tool is strongly based upon the previous research and work put together by
# kingshriek, the Shenmue team (especially Sappharad and LemonHaze), and many others.
# All wisdom is from them, all mistakes are from me.

# In the interest of clear nomenclature, I'm going with track->sample(s)

# Update history located near end of file

# TODO list:
# * bugbug/todos as noted in the code
# * add BLANK option?
# * add delete option?
# * add joined-sample option?
# * add add-sample option?
# * add resampling to help with size concerns?
# * add volume change?
# * add rate change for samples reused at different rates (such as mvc2 Tron Bonne)?

# MIT License

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# =====================================================================================================

# =====================================================================================================
import os
import sys
import wave
import math
from struct import pack
from struct import unpack
from array import array
from enum import Enum
# =====================================================================================================

# =====================================================================================================

# Config options since there's a lot of data
show_detailed_file_info = False
show_detailed_data_tracks = False
show_detailed_data_playback = False
show_detailed_data_samples = False
show_detailed_data_samples_full = False
dump_adpcm_samples = False
dump_pcm_samples_to_wav = False
dump_adpcm_samples_to_wav = False
# FWIW, 6 * 156000 = 936000.  The MvC2 common match effects file (SE_COMN.BIN) is 931284.
# They apparently worked hard to fit within the 2MB audio limit for Dreamcast, and if you're 
# injecting new samples you would want to adhere to that same minimalist approach.
mvc2_DTPK_size_limit = 156000

INVALID_SAMPLE_ID = 0xFFFF

# =====================================================================================================
# I'm unclear where the bcolors implementation sources from: it's referenced around the internet.
# Usage is as shown: make sure to close the presentation shift with ENDC.

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def PrintAsWarning(msg):
	print (bcolors.WARNING + msg + bcolors.ENDC)

def PrintAsFailure(msg):
	print (bcolors.FAIL + msg + bcolors.ENDC)

# =====================================================================================================
def DelayValuesTo44100SampleCount(unMaskedByte1, unMaskedByte2):
	# For a discussion of what these magic values are, see where we ingest the track composition data
	# Byte 1:
		# starts masked with 0x80
		# each odd +0x01 step is +0n5120 samples
		# each even +0x01 step is +0n6144 samples
		# a full 0x10 steps is +0n90112 samples (which is (8 * 5120) + (8 * 6144) conveniently enough)
	# Byte 2 is optional, and increments at chunks of 1024 samples every 0x17 steps
	delayInSamples = (int(math.floor(unMaskedByte1 / 2)) * 11264) + ((unMaskedByte1 % 2) * 5120)
	delayInSamples += int(math.floor(unMaskedByte2 / 23)) * 1024
	return delayInSamples

# =====================================================================================================
class DTPKTrackEntry:
	# A normal track entry looks like:
	# C0 DF 01 78 80 FF
	# In order:
	# Start_Track (C0) Sample_Type (DF) Sample_ID (01) Volume (78) End_Sample_Ref? (80) End_Track (FF)
	# Note: Maximum value of volume is 0x7f
	# For a Joined track such as Track 01 in SND_OPEN.BIN for cvs2:
	# C0 DF 01 78 00 8B 4E DF 02 78 80 FF
	# In order:
	# Start_Track (C0) Sample_Type (DF) Sample_ID (01) Volume (78) Join_With (00) Delay_Part1 (0b | 80) Opt_Delay_Part2 (4e)
	#    Sample_Type (DF) Sample_ID (02) Volume (78) End_Sample_Ref (80) End_Track (FF)
	# The Delay value is relative to 0:00, not the end of the previous sample.

	def __init__(self, entry_type, sample_id, volume, trailing_bytes):
		self.entry_type = entry_type
		self.sample_id = sample_id
		self.volume = volume
		self.trailing_bytes = trailing_bytes

	def Print(self):
		if (IsTrackEntryUsed(self.entry_type)):
			summary = '\t\t\t\tType: 0x%02x.  Sample: 0x%02x.  Volume?: 0x%02x.' % (self.entry_type, self.sample_id, self.volume)

			extraBytes = ""

			# Track Entry errata:
			# Cammy-CVS2 (SND_PL0C) track 36 "joins" two non-existent samples (25, 26) using terminal byte(?) 80
			# not a join/append byte (00/01).  But it's an unused track, so the composition is irrelevant.
			if (self.trailing_bytes[0] & 0x80):
				extraBytes = ' Terminal byte: 0x%02x' % self.trailing_bytes[0]
			else:
				extraBytes = " Join bytes:"
				for val in self.trailing_bytes:
					extraBytes += " 0x%02x" % val

				if (self.trailing_bytes[0] == 0):
					extraBytes += '\n\t\t\t\t    Meaning: Join Next (0x%02x)' % self.trailing_bytes[0]
				else:
					extraBytes += '\n\t\t\t\t    Meaning: Replace With? (0x%02x)' % self.trailing_bytes[0]

				delayInSecs = 0
				if (len(self.trailing_bytes) > 1):
					unMaskedByte1 = self.trailing_bytes[1] ^ 0x80
					unMaskedByte2 = 0
					extraBytes += ' Delay Next: 0x%02x' % unMaskedByte1
					if (len(self.trailing_bytes) > 2):
						unMaskedByte2 = self.trailing_bytes[2]
						extraBytes += ' 0x%02x' % self.trailing_bytes[2]
					delayInSamples = DelayValuesTo44100SampleCount(unMaskedByte1, unMaskedByte2)
					extraBytes += ' (~%1.3fs)' % (delayInSamples / 44100) # I did all my math at 44100
				if (len(self.trailing_bytes) > 3):
					PrintAsFailure('ERROR: Unknown track entry data {}'.format(self.trailing_bytes))

			summary += extraBytes

		elif (self.entry_type == 0xa0):
			# Unused: our internal record of sample_id is INVALID_SAMPLE_ID here, and we stuff the unknown byte into Volume
			summary = '\t\t\t\tUNUSED: Type: 0x%02x.  Sample: N/A (0x%02x).  Terminal byte: 0x%02x' % (self.entry_type, self.volume, self.trailing_bytes[0])
		else:
			summary = '\t\t\t' + bcolors.FAIL + 'ERROR: Unknown type: 0x%02x.  Sample: N/A (0x%02x).  Terminal byte: 0x%02x' % (self.entry_type, self.sample_id, self.trailing_bytes[0]) + bcolors.ENDC
		print (summary)


class DTPKTrack:
	def __init__(self, track_id, playback_id):
		self.track_id = track_id
		self.playback_id = playback_id
		self.entryList = []

	def AddSample(self, entry_type, sample_used, volume, trailing_bytes):
		self.entryList.append(DTPKTrackEntry(entry_type, sample_used, volume, trailing_bytes))

	def GetSampleCount(self):
		sampleCount = 0
		for entry in self.entryList:
			if (entry.sample_id != INVALID_SAMPLE_ID):
				sampleCount += 1
		return sampleCount

	def Print(self):
		if (show_detailed_data_tracks):
			if (IsTrackEntryUsed(self.entryList[0].entry_type)):
				# At least one
				print ('\t\t\tTrack %02u composition (SPD 0x%02x), using %u samples:' % (self.track_id, self.playback_id, len(self.entryList)))
			else:
				# Unused
				print ('\t\t\tTrack %02u composition (Unused):' % self.track_id)
			for entry in self.entryList:
				entry.Print()

class DTPKPlaybackDef:
	def __init__(self, playback_id, track_loc, sample_rate, sample_used):
		self.playback_id = playback_id
		self.track_loc = track_loc
		self.rate = sample_rate
		self.sample_used = sample_used

	def Print(self):
		if (show_detailed_data_playback):
			print ('\t\tPlayback data %02u (0x%02x) at 0x%04x.  Rate: 0x%02x (%5uHz).  Uses sample %02u (0x%02x)' % (self.playback_id, self.playback_id, self.track_loc, self.rate, TranslateDTPKRate(self.rate), self.sample_used, self.sample_used))

class DTPKSample:
	def __init__(self, sample_id, sample_def_offset, sample_flags, sample_loop_start, sample_loop_end, sample_rate, sample_length, group_id, track_id, playback_id, real_rate):
		self.id = sample_id
		self.def_offset = sample_def_offset

		self.data_offset = sample_flags & 0x007fffff
		self.quality     = sample_flags & 0x00800000
		self.format      = sample_flags & 0x01000000
		self.unk         = sample_flags & 0x02000000
		self.loop_start = sample_loop_start
		self.loop_end = sample_loop_end
		# This is unused in our examples
		self.rate = sample_rate
		self.length = sample_length

		# These are grabbed from other chunks and are not present in the sample data
		self.group_id = group_id
		self.track_id = track_id
		self.playback_id = playback_id
		self.real_rate = real_rate

	def GetFlagValueForNewLocation(self, location):
			if ((self.format == 0x1000000) and (self.quality == 0)):
				return location | 0x1000000 | 0x2000000
			elif ((self.format == 0) and (self.quality == 0)):
				return location
			elif ((self.format == 0) and (self.quality == 0x800000)):
				return location | 0x0800000

	def Print(self):
		if (show_detailed_data_samples_full):
			print('\t\tSample %2u / 0x%02x (def at 0x%x)' % (self.id, self.id, self.def_offset))
			print('\t\t\tLocation: 0x%05x' % self.data_offset)

			if ((self.format == 0x1000000) and (self.quality == 0)):
				print ('\t\t\tsample_format:  ADPCM  (0x1000000)')
				print ('\t\t\tsample_format:  4bit   (0x0000000)')
			elif ((self.format == 0) and (self.quality == 0)):
				print ('\t\t\tsample_format:  PCM    (0x0000000)')
				print ('\t\t\tsample_quality: 16bit  (0x0000000)')
			elif ((self.format == 0) and (self.quality == 0x800000)):
				print ('\t\t\tsample_format:  PCM    (0x0000000)')
				print ('\t\t\tsample_quality: 8bit   (0x0800000)')
			else:
				PrintAsFailure('ERROR!!! This is not a valid audio format for the AICA')
				print ('\t\t\tsample_format:  Unknown (0x%07x)', self.format)
				print ('\t\t\tsample_quality: Unknown (0x%07x)', self.quality)

			print ('\t\t\tsample_unk:             0x%07x' % self.unk)

			print ('\t\t\tsample_loop_start:      0x%04x count samples' % self.loop_start)
			print ('\t\t\tsample_loop_end:        0x%04x count samples' % self.loop_end)
			print ('\t\t\tsample_rate:            0x%x' % self.rate)
			print ('\t\t\tsample_length:          0x%08x bytes' % self.length)
			print ('\t\t\t\t(sample data is at 0x%x to 0x%x)' % (self.data_offset, self.data_offset + self.length))
		elif (show_detailed_data_samples):
			print ('\t\tSample %02u (def at 0x%x)' % (self.id, self.def_offset))

	def GetFilename(self, filename_base, extension):
		if (self.group_id == 0xffff):
			strGroupId = "NA"
		else:
			strGroupId = '%02u' % self.group_id

		if (self.track_id == 0xffff):
			strTrackId = "NA"
		else:
			strTrackId = '%02u' % self.track_id

		if (self.playback_id == 0xffff):
			strPlaybackId = "NA"
		else:
			strPlaybackId = '%02u' % self.playback_id

		#create file
		# bugbug todo: It may be much more useful to have the frequency in the filename rather than the internal rate value
		if (self.rate == 0):
			filename_out = ('%s_%s_%s_SPD_%s_Sample_%02X_Rate_%02X.%s' % (filename_base, strGroupId, strTrackId, strPlaybackId, self.id, self.real_rate, extension))
		else:
			filename_out = ('%s_%s_%s_SPD_%s_Sample_%02X_Rate_%02X.%s' % (filename_base, strGroupId, strTrackId, strPlaybackId, self.id, self.rate, extension))

		return filename_out

	# This needed utility code comes from hl037_ on stackoverflow
	def signedToUnsigned(self, n, byte_count): 
		return int.from_bytes(n.to_bytes(byte_count, 'little', signed=True), 'little', signed=False)

# =====================================================================================================
# This ADPCM->PCM region is thanks to Kode54, BERO, Megan Potter, and Sappharad.
# Sappharad's work is under the MIT License as included above.
# BERO is under GPL.
# Kode54 is under MAME GPL v2 or similar
# I'm unclear as to the licensing for KallistiOS and/or Megan Potter's work.
# This section is entirely a straight Python-ization of Sappharad's C# adpcm2wav code
# There is no new value or knowledge added here, it's just going to produce byte-identical
# output to that C# version.

#region AICA ADPCM decoding
	def adpcm2pcm(self, file_data, src, length):
		diff_lookup = array('i', [ 1, 3, 5, 7, 9, 11, 13, 15, -1, -3, -5, -7, -9, -11, -13, -15 ])
		index_scale = array('i', [ 0x0e6, 0x0e6, 0x0e6, 0x0e6, 0x133, 0x199, 0x200, 0x266 ])

		dst = bytearray(length * 4)
		dstLoc = 0
		cur_quant = 0x7f
		cur_sample = 0
		highNybble = False

		while (dstLoc < len(dst)):
			if (highNybble):
				shift1 = 4
			else:
				shift1 = 0
			delta = (file_data[src] >> shift1) & 0xf

			# ~2014 mame version by Kode54 / Metallic.  In testing Sapphard's appears to be more accurate for our needs.
			# Note that if you use this, get rid of the 254/256 line below
			#x = (cur_quant * diff_lookup[delta & 7]) / 8
			#x = min(x, 0x7fff)
			#if (delta & 8):
			#	x = -x

			x = cur_quant * diff_lookup[delta & 15]
			x = int(cur_sample + ((x + (self.signedToUnsigned(x, 4) >> 29)) >> 3))


			# standard cleanup
			cur_sample = min(max(x, -32768), 32767)
			cur_quant = (cur_quant * index_scale[delta & 7]) >> 8
			cur_quant = min(max(cur_quant, 0x7F), 0x6000)

			# standard byte packing
			dst[dstLoc] = (int(cur_sample) & 0xFF)
			dstLoc += 1
			dst[dstLoc] = ((int(cur_sample) >> 8) & 0xFF)
			dstLoc += 1

			cur_sample = int(cur_sample * 254 / 256)

			highNybble = not highNybble
			if (not highNybble):
				src += 1

		return dst
	
#endregion


	def ExtractIfYADPCM(self, filename_base, file_data):
		if ((self.format == 0x1000000) and (self.quality == 0)):
			filename_out = self.GetFilename(filename_base, "yadpcm")
			sample_data = file_data[self.data_offset:self.data_offset + self.length]

			try:
				open(filename_out, 'wb').write(sample_data)
				print ('\tADPCM export: %s' % filename_out)
				return 1
			except:
				PrintAsFailure('ERROR: filename \'%s\' cannot be written to currently.' % filename_out)

			return 1
		return 0

	def ExtractPCMToWAV(self, filename_base, file_data):
		# Note that this sketches in nonPCM export, but that won't work because we don't have a VFW decoder
		# The Shenmue team at https://github.com/Shenmue-Mods/DTPKUtil/blob/master/DTPKutil/DtpkFile.cs
		# have a fairly functional adpcmTopcm converter, but it does skewed conversion for 
		# our files
		# VGMSteam's code at https://github.com/vgmstream/vgmstream/blob/master/src/coding/yamaha_decoder.c
		# is also interesting
		if (self.format == 0x0000000) or (dump_adpcm_samples_to_wav):
			filename_out = self.GetFilename(filename_base, "wav")
		
			riffHeader = b'RIFF'
			riffType = b'WAVE'
			chunkFmtName = b'fmt '

			# 2b formattag
			# So uh MSFT's formattag registration tables are a little screwy.  In theory
			# the correct formattag might be 0n20 / 0x14 for Yamaha ADPCM, but I haven't
			# seen a VFW decoder that works so who actually knows.  But hey: use 0n20 if 
			# want to be exporting ADPCM WAV
			# PCM == 0x01
			wFormatTag = b'\x01\x00'

			# 2b number of channels
			# ... everything shall be mono here
			nChannels = b'\x01\x00'

			# 4b sample rate: use the DTPK rate value
			# 4b bytes per second: likewise
			rate_to_use = self.rate
			if (rate_to_use == 0):
				rate_to_use = self.real_rate

			frequency = TranslateDTPKRate(rate_to_use)
			nSamplesPerSec = nAvgBytesPerSec = frequency.to_bytes(4, byteorder='little')

			# 2b block align
			nBlockAlign = b'\x02\x00'
			# 2b bits per sample

			fIsHalfQuality = False
			if ((self.format == 0x1000000) and (self.quality == 0)):
				# We convert ADPCM 4bit to PCM 16bit for export
				nBitsPerSample = b'\x10\x00'
			elif ((self.format == 0) and (self.quality == 0)):
				# PCM 16bit
				nBitsPerSample = b'\x10\x00'
			elif ((self.format == 0) and (self.quality == 0x800000)):
				# PCM 8bit
				nBitsPerSample = b'\x08\x00'
				fIsHalfQuality = True

			exData = b''

			# You only need this if you're exporting out ADPCM-as-ADPCM within WAV.  But again: no VFW decoder
			# for that that I'm aware of
			#if (self.format == 0x1000000): #ADPCM
			#	exData = b'\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' 

			chunkFmtData = wFormatTag + nChannels + nSamplesPerSec + nAvgBytesPerSec + nBlockAlign + nBitsPerSample + exData

			chunkFmtSize = pack('l', len(chunkFmtData))

			chunkDataName = b'data'

			fWasYADPCM = False
			if (self.format == 0x1000000):
				fWasYADPCM = True
				chunkDataData = self.adpcm2pcm(file_data, self.data_offset, self.length)
			else:
				chunkDataData = file_data[self.data_offset:self.data_offset + self.length]

			if fIsHalfQuality:
				# Xor the data here and I don't know why
				for iPos, iCurr in enumerate(chunkDataData):
					chunkDataData[iPos] = chunkDataData[iPos] ^ 0x80

			chunkDataSize = pack('l',len(chunkDataData))

			riffBody = riffType + chunkFmtName + chunkFmtSize + chunkFmtData + chunkDataName + chunkDataSize + chunkDataData

			fileSize = pack('l', len(riffBody))

			try:
				open(filename_out, 'wb').write(riffHeader + fileSize + riffBody)
				if (fWasYADPCM):
					print ('\tPCM export: %s\t(from ADPCM)' % filename_out)
				else:
					print ('\tPCM export: %s' % filename_out)
				return 1
			except:
				PrintAsFailure('ERROR: filename \'%s\' cannot be written to currently.' % filename_out)
		return 0
		
# =====================================================================================================
def IsTrackEntryStart(entryValue):
	if ((entryValue == 0xd1) or # seen in doa2
		(entryValue == 0xd2) or # seen in doa2
		(entryValue == 0xc0)):  # seen in cvs2/vtennis/mvc2
		return True
	return False

# =====================================================================================================
def IsTrackEntryUsed(entryType):
	if ((entryType == 0xdc) or # seen in cvs2
	    (entryType == 0xdd) or # seen in cvs2
	    (entryType == 0xde) or # seen in cvs2
		(entryType == 0xdf)):  # seen in mvc2
		return True
	return False

# =====================================================================================================
def TranslateDTPKRate(frequency):
	#DTPK rate values can be found in these examples:
	# Unknown values (bugbug todo)
	# 01 doa fight_se
		#  78/0x4e, sample 0x23 - track 17 join?
		# 120/0x78, sample 0x21 - track 62 join?
		# 123/0x7b, sample 0x16 - track 73 join?
		# 148/0x94, sample 0x29 - track 78 join?
	# 02 (doa2 fight_se km_c2/ks_c2/kt_end/tg_c123): ?  if this is correct, this is unknown.
		# 25/0x19, sample 0x28 - part of track 8
		# 74/0x4a sample 0x29
	# 04 cvs2 SND_PL2A (Hibiki): track 49/0x31, sample 39/0x27, also track 103/0x67
		# confirmed in file
		# seems to be about 44100-45000?
	# 05 doa fight_se
		# 75/0x4b, sample 0x29

	# Unused value...?  Not going to make this the new zero_value because if there is usage of 0xd1-0xdd I'd like to test them first
	# d1 DOA voice_le voice_lu: unused - empty samples

	# "Known" values:
	# 0x0 44100: doa2
	
	# 0xde 6000: mvc2 pl3a kobun/servbot SPD 0x1f sample 0x1a track 39 (slowdown of 38)
	# 0xdf 6500: unused? not seen in content yet
	# 0xe0 7000: unused? not seen in content yet
	# 0xe1 7500: mvc2 pl10_voi tron bonne, spd 0x17, 0x18, 0x41, 0x42 - sample 0x20 0x15, track 30 (slowdown of 29) and 32
	# 0xe2 8000: mvc2 guile 02
	# 0xe3 8500: speculative: needs verification
	# 0xe4 9000: speculative: seems good for mvc2 sakura, abyss
	# 0xe5 9500: speculative: needs verification
	# 0xe6 10000: cvs2 rolento
	# 0xe7 10500: speculative: seems good for mvc2 sakura
	# 0xe8 11025: most
	# 0xe9 12000: cvs2 rolento
	# 0xea 12500: speculative: needs verification
	# 0xeb 13000: speculative: seems good for mvc2 bonerine
	# 0xec 14000: speculative: needs verification
	# 0xed 15000: speculative: needs verification
	# 0xee 16000: speculative: seems good for mvc2 storm, jin
	# 0xef 17000: speculative: seems good for doa2 voice_ba
	# 0xf0 18000: speculative: seems good for mvc2 storm, spiral
	# 0xf1 19000: speculative: seems good for doa2 fight_se/voice_kl, cvs2 guile
	# 0xf2 20000: speculative: seems good for vtennis pack 8
	# 0xf3 21000: speculative: seems good for vtennis pack 8
	# 0xf4 22050: lots
	# 0xf5 23000: cvs2 dhalsim
	# 0xf6 24000: cvs2 vega / vtennis pack 4
	# 0xf7 26000: speculative: seems good for vtennis pack4, doa2 km_c1
	# 0xf8 28000: speculative: seems good for doa2 fight_se
	# 0xf9 30000: speculative: needs verification
	# 0xfa 32000: cvs2 cammy
	# 0xfb 34000: virt tennis pack0
	# 0xfc 35000: speculative: needs verification
	# 0xfd 38000: speculative: seems good for vtennis pack0
	# 0xfe 40000: speculative: seems good for doa2 fight_se
	# 0xff 42000: speculative: seems good for doa2 fight_se

	rateMap = [ 6000, 6500, 7000, 7500, 8000, # de...
				8500, 9000, 9500, 10000, 10500, 11025, # e3...
				12000, 12500, 13000, 14000, 15000, 16000, 17000, # e9...
				18000, 19000, 20000, 21000, 22050, 23000, 24000, # f0...
				26000, 28000, 30000, 32000, 34000, 35000, 38000, # f7...
				40000, 42000, # 0xfe 0xff
				# Note that everything above 44100 is highly speculative
				44100, 45000, 46000, 47000, 48000, 49000 # 0x00 0x01 0x02 0x03 0x04 0x05
				]

	# This is lowest_known_DTPK_rate_value
	zero_val = 0xde
	max_low_value = 0x05

	if (frequency < 0x100): # DTPK rate to frequency
		table_index = frequency
		if ((table_index > max_low_value) and (table_index < zero_val)):
			if (table_index == 0xd1):
				PrintAsWarning('Errata: unusual rate 0xd1, seems only to be used for empty samples.  Possibly 0hz?')
			else:
				PrintAsFailure('ERROR: unknown DTPK rate value 0x%02x: Please let Preppy know about this file so we can figure it out.' % frequency)
			return 22050
		if (table_index <= max_low_value):
			table_index += 0x100

		table_index -= zero_val

		return rateMap[table_index]
	else: # frequency to DTPK rate
		for iPos, iVal in enumerate(rateMap):
			if (iVal == frequency):
				return (iPos + zero_val) & 0xFF
		PrintAsWarning('Warning: frequency %u is not supported in DTPK.  For best results please resample this to a supported rate' % frequency)
		# No exact match: use a close-ish one
		for iPos, iVal in enumerate(rateMap):
			if (iPos == 0):
				continue
			if (frequency <= iVal):
				return iVal
		return 0
	
# =====================================================================================================
class InjectableSampleType(Enum):
	PCM16  = 1
	PCM8   = 2
	ADPCM4 = 3

# =====================================================================================================
class InjectableSample:
	samplePosition = 0
	sampleType = InjectableSampleType.PCM16
	rate = 22050
	dataObject = []

	def __init__(self, position, type, waveData, rate):
		self.samplePosition = position
		self.sampleType = type
		self.dataObject = waveData

		# If we get a DTPK rate value, convert it to the standard playback rate
		if (rate < 0x100):
			rate = TranslateDTPKRate(rate)

		self.rate = rate

	def GetHeader(self, location):
		# Rate seems to be unused in this data chunk
		sampleRate = 0
		sampleLength = len(self.dataObject)
		if (self.sampleType == InjectableSampleType.ADPCM4):
			flags = location | 0x1000000 | 0x02000000
			loopEnd = sampleLength * 2
			# 28 is a speculative magic number here as seen in cvs2 adpcm samples
			# and as such could be wrong
			loopStart = loopEnd - 28
		elif (self.sampleType == InjectableSampleType.PCM8):
			flags = location | 0x0800000
			loopStart = 0
			loopEnd = sampleLength
		elif (self.sampleType == InjectableSampleType.PCM16):
			flags = location
			loopStart = 0
			loopEnd = int(sampleLength / 2)

		outBytes = bytearray(16)
		outBytes[0:4] = array('B', flags.to_bytes(4, byteorder='little'))
		outBytes[4:6] = array('B', loopStart.to_bytes(2, byteorder='little'))
		outBytes[6:8] = array('B', loopEnd.to_bytes(2, byteorder='little'))
		outBytes[8:12] = array('B', b'\x00\x00\x00\x00')
		outBytes[12:16] = array('B', sampleLength.to_bytes(4, byteorder='little'))
		return outBytes

# =====================================================================================================
def ImportAndAddADPCMFile(adpcmFilename, sampleList, samplePosition, sampleRate):
	# This is just a raw data chunk
	dataChunk = array('B', open(adpcmFilename, 'rb').read())
	sampleType = InjectableSampleType.ADPCM4
	sampleList.append(InjectableSample(samplePosition, sampleType, dataChunk, sampleRate))
	return True

# =====================================================================================================
def ImportAndAddWAVFile(wavFilename, sampleList, samplePosition):
	try:
		waveObject = wave.open(wavFilename, 'rb')
	except:
		PrintAsFailure("Not a supported WAV: only raw PCM is natively supported in Python")
		return False

	isValid = True
	try:
		nChannel = waveObject.getnchannels()
		if (nChannel != 1):
			print ('Only mono WAV is supported')
			raise

		nSampleWidth = waveObject.getsampwidth()
		if (nSampleWidth == 1):
			sampleType = InjectableSampleType.PCM8
		elif (nSampleWidth == 2):
			sampleType = InjectableSampleType.PCM16
		else:
			PrintAsFailure('ERROR: sample width is {}: only 1 or 2 is supported'.format(nSampleWidth))
			raise

		nFrameRate = waveObject.getframerate()
		if (TranslateDTPKRate(nFrameRate) == 0xbadf00d):
			PrintAsFailure('ERROR: framerate is {}: only a few rates are currently supported, check code'.format(nFrameRate))
			raise

		nFrames = waveObject.getnframes()

		flDuration = nFrames / nFrameRate

		dataChunk = waveObject.readframes(nFrames)

		if (nSampleWidth == 1):
			# Unknown XOR as noted below
			fixedAudio = bytearray(len(dataChunk))
			for iPos, iCurr in enumerate(dataChunk):
				fixedAudio[iPos] = dataChunk[iPos] ^ 0x80
			dataChunk = fixedAudio

		print ('Requesting to inject file {}:\n\tChannels {} SampWidth {} Framerate {} Count frames {} Duration {:2.2}s\n'.format(wavFilename, nChannel, nSampleWidth, nFrameRate, nFrames, flDuration) )

		sampleList.append(InjectableSample(samplePosition, sampleType, dataChunk, nFrameRate))

	except:
		PrintAsFailure('ERROR: Unable to load this WAV file.')
		isValid = False

	waveObject.close()

	return isValid


# =====================================================================================================


# =====================================================================================================
# Display data in the DTPK file
def dtpkparse(dtpkfile, outFile, changedSamples):
	try:
		dtpkbytes = array('B', open(dtpkfile, 'rb').read())
	except:
		PrintAsFailure('ERROR: \'%s\' could not be opened.' % dtpkfile)
		sys.exit(0)

	fIsProbablyMvC2CharacterFile =  (dtpkfile.upper().startswith("PL") and dtpkfile.upper().endswith("_VOI.BIN"))

	if (dtpkbytes[:4].tobytes() != b'DTPK'):
		print ('Skipping %s - not a valid DTPK sound file' % dtpkfile )
		return

	(filename_base, filename_ext) = os.path.splitext(dtpkfile)

	fIsShowingAnyInfo = show_detailed_file_info or show_detailed_data_tracks or show_detailed_data_playback or show_detailed_data_samples or show_detailed_data_samples_full

	dtpk_id = unpack('<I', dtpkbytes[4:8])[0]
	# Note we may adjust this size for write-out if we splice in samples
	filesize = unpack('<I', dtpkbytes[8:12])[0]
	filesize_actual = len(dtpkbytes)
	if (filesize != filesize_actual):
		# This will cause unexpected errors, so mention it to the user.
		PrintAsFailure('ERROR! Filesize in file is listed as 0x%x, but actual filesize is 0x%x.  This file may not play back correctly.' % (filesize, filesize_actual))

	seq_data_loc = unpack('<I', dtpkbytes[0x2C:0x30])[0]
	if not seq_data_loc:
		print ('Skipping %s - no sequence data found' % dtpkfile)
		return

	sample_playback_def_header_loc = unpack('<I', dtpkbytes[0x30:0x34])[0]

	sample_offset = unpack('<I', dtpkbytes[0x3C:0x40])[0]
	ncmds = unpack('<I', dtpkbytes[seq_data_loc:seq_data_loc + 4])[0] + 1

	seq_cmds = []
	seqcmds_loc = seq_data_loc + 4

	if (show_detailed_file_info):
		print ('Analyzing %s:' % dtpkfile)
		print ('\tSignature:\t\tDTPK\t\t\t0x0')
		print ('\tDTPK id:\t\t0x%x\t\t\t0x4' % dtpk_id)
		print ('\tFilesize:\t\t0x%x (0n%u)\t0x8' % (filesize, filesize))
		print ('\tSeq data starts at:\t0x%x\t\t\t0x2c' % seq_data_loc)
		print ('\tSample defs starts at:\t0x%x\t\t\t0x30' % sample_playback_def_header_loc)
		print ('\tSample data starts at:\t0x%x\t\t\t0x3c' % sample_offset)
		print ('\tSeq group count:\t%u\t\t\t0x%x' % (ncmds, seq_data_loc))

	current_group = 1
	for icmd in range(ncmds):
		cmd_data = unpack('<I', dtpkbytes[seqcmds_loc:seqcmds_loc + 4])[0]
		cmd_offset = cmd_data & 0xFFFF
		cmd_type = (cmd_data & 0xFFFF0000) >> 16

		if (show_detailed_file_info):
			print ('\tGroup %u description:\t0x%x\t\t\t0x%x' % (current_group, cmd_type, seqcmds_loc))
			if (cmd_type >> 8) == 0xA8:
				print ('\t\t\t\t\tA8: Music')
			elif (cmd_type >> 8) == 0xA9:
				print ('\t\t\t\t\tA9: Voice data')
			elif (cmd_type >> 8) == 0xAA:
				print ('\t\t\t\t\tAA: Sound effects/other')
			else:
				# DOA2 uses AB for example
				print ('\t\t\t\t\t(Unknown group type)')

		seqcmds_loc += 4
		current_group += 1

	# Get number of tracks
	ntracks = unpack('<I', dtpkbytes[seqcmds_loc:seqcmds_loc + 4])[0] + 1

	if (show_detailed_file_info):
		print ('\tCount tracks:\t\t%u (0x%x)\t\t\t0x%x' % (ntracks, ntracks, seqcmds_loc))

	track_offset = seqcmds_loc + 4

	# Now we're into speculatively understood data chunks that we're ironing out details on. 
	# Clarifications always appreciated.  In my 44 track sample (PL0A_VOI 'rogue' from mvc2-dc):
	# Sequencer 1 data starts after track count at 0x1b4:
		# Track Composition Read Offset data group 1 from 0x1b4-0x268
			# 4 bytes per track.
			# First TCRO entry is the start_byte for all TCD entries
			# Every additional entry represents the byte read offset within the TCD for that specific track OR'd with the Start_Byte value
			# Consider:
			# c0 00 00 00 c6 00 00 00 cc 00 00 00
			# This indicates the Start_Byte for TCD entries shall be 0xc0.
			# Track 0 starts at offset 0 within the TCD.  Track 1 starts at offset 0x06.  Track 2 starts at offset 0x0c.
		# Track Composition Data from 0x268-0x358 // len 0xf0 / 240 .  Variable length per track, each section ends with FF
	# We are now in data for sequencer group 2 as of 0x358:
		# Track Composition Read Offset data group 1 0x358-0x40c // len b4 / 180
		# Track Composition Data from 0x40c-0x4fc // 0xf0 // 240
	# Unknown data chunk from 0x4fc-0x54c // 0x50 // 80
	# Following this, we have Track Playback Data from 0x54c-0xf4c in 64byte chunks for those 44 tracks
		# That is length 0xa00  / 0n2560
	count_samples = unpack('<I', dtpkbytes[sample_offset:sample_offset + 4])[0] + 1
	if (show_detailed_file_info):
		print ('\tSample count:\t\t%u (0x%x)\t\t\t0x%x' % (count_samples, count_samples, sample_offset))

	sample_offset = sample_offset + 4

	# Now we have N sets of tables based upon how many sequencer groups there are.
	sequencer_data_tables_offset = seq_data_loc
	# First at 0x1a8 is an 4 byte header + 4 bytes per sequencer group
	sequencer_data_tables_offset += 4 + (4 * ncmds)

	allGroups = []

	if (show_detailed_file_info):
		print ('\tSequencer group data (0x%03x):' % sequencer_data_tables_offset)

	# OK, so each Group can have a variable number of Tracks capped by the MaxTrack value.
	# Each used Track is associated with a unique sample playback definition in that table.
	nTotalSPDsUsed = 0

	for iSeqGroup in range(ncmds):
		track_composition_entry_offset_table_loc = sequencer_data_tables_offset
		track_composition_entry_count = unpack('<I', dtpkbytes[track_composition_entry_offset_table_loc:track_composition_entry_offset_table_loc + 4])[0] + 1
		track_composition_entry_offset_table_length = (track_composition_entry_count * 0x04) + 0x04
		track_composition_table_loc = track_composition_entry_offset_table_loc + track_composition_entry_offset_table_length

		allTracks = []

		# Table 1:
		# 4 byte header.  First byte is track count
		# 4 bytes per track.  First two bytes are number of bytes offset from the sequencer table start for that track
		# TCD reads themselves are unbounded until they hit the terminal TCD bytes (0x80 0xFF)
		offsetEntryTableLoc = track_composition_entry_offset_table_loc + 4 # Advance past count
		if (show_detailed_file_info or show_detailed_data_tracks):
			print ('\t\tGroup %u track composition offset data (starts at 0x%x)' % (iSeqGroup, offsetEntryTableLoc))

			for iCurEntry in range(track_composition_entry_count):
				positionWithinSequencerTable = unpack('<I', dtpkbytes[offsetEntryTableLoc:offsetEntryTableLoc + 4])[0]
				if (show_detailed_file_info or show_detailed_data_tracks):
					print ('\t\t\tTrack data %2u (0x%x): Definition starts at 0x%x (0x%x from sequencer table start)' % (iCurEntry, offsetEntryTableLoc, seq_data_loc + positionWithinSequencerTable, positionWithinSequencerTable))
				offsetEntryTableLoc += 4

		# Table 2: 5-8 bytes per track.
			# ~5 bytes per C0 A0 entry
			# ~6 bytes per C0 DF entry
			# A0 entries appear to indicate unused tracks
			# Terminal bytes appear to be 0xff80 OR 0xff81 (ff81 in sound effects files only so far)
			# This table can be used to join samples together!
			# For joins, based upon ~350 files, there is either two or three join bytes.
			# Consider AA BB [CC].
			# AA is wholly unknown but is either 00 (most often) or 01 (rarely)
			# BB appears to be 80 OR'd with a delay duration in 0.25s increments(?).  maybe in samples? unclear
			# CC is often present and seems to increase the delay duration
			# OK, spent some time generating variant DTPKs with different join delay values, and the values appear to be:
			# BB ^ 80 = variable length step.
				# each odd step is +0n5120 samples
				# each even step is +0n6144 samples
				# a full 0x10 steps is +0n90112 samples.
			# CC = floor(CC/0n23) * 0n1024 samples, so about 0.023s per increment
			# That's all speculative based upon an analysis of about 90 variant delays from 8000 (play at same time) to b96e, done with
			# both mvc2-dc and cvs2-dc files.
			# Somewhere between 0xb96e to 0xc96e we ... overflow(?) and the second/joined clip just doesn't play.
		unk_track_loc = track_composition_table_loc

#		if (show_detailed_file_info):
#			print ('\t\tUnknown data (0x%x - 0x%x): 0x%x bytes' % (track_composition_entry_offset_table_loc, track_composition_table_loc, track_composition_entry_offset_table_length))
		if (show_detailed_file_info or show_detailed_data_tracks):
			print ('\t\tGroup %u track composition data (starts at 0x%x):' % (iSeqGroup, track_composition_table_loc))

		for iUnkTrack in range(track_composition_entry_count):
			if (IsTrackEntryStart((dtpkbytes[unk_track_loc:unk_track_loc+1])[0])):
				# entry
				unk_track_loc += 1
				entry_type = (dtpkbytes[unk_track_loc:unk_track_loc+1])[0]

				current_track = DTPKTrack(iUnkTrack, nTotalSPDsUsed)

				if (IsTrackEntryUsed(entry_type)):
					while (IsTrackEntryUsed(entry_type)):
						trailing_bytes = []
						nTotalSPDsUsed += 1
						unk_track_loc += 1
						# sample to use
						sample_to_use = (dtpkbytes[unk_track_loc:unk_track_loc+1])[0]
						unk_track_loc += 1

						# Note that there can be a conflict(?) between the Track Composition Data sample value and 
						# the Track Playback Data sample value.  For at least flycast, the non-zero value of the two is ignored.
						# In the instance where both are non-zero, the Track Composition Data value is used.
						volume_maybe = (dtpkbytes[unk_track_loc:unk_track_loc+1])[0]
						unk_track_loc += 1
						
						while (((dtpkbytes[unk_track_loc:unk_track_loc+1])[0] != 0xa0) and
							   (not (IsTrackEntryUsed((dtpkbytes[unk_track_loc:unk_track_loc+1])[0]))) and 
							   ((dtpkbytes[unk_track_loc:unk_track_loc+1])[0] != 0xff)):
							# optional bytes
							cur_byte = (dtpkbytes[unk_track_loc:unk_track_loc+1])[0]
							# first byte in the segment appears to be volume...?
							trailing_bytes.append(cur_byte)
							unk_track_loc += 1
							# closure or join byte is 0xff
						#trailing_bytes = "%s." % (trailing_bytes)
						current_track.AddSample(entry_type, sample_to_use, volume_maybe, trailing_bytes)
						entry_type = (dtpkbytes[unk_track_loc:unk_track_loc+1])[0]
				elif (entry_type == 0xa0):
					unk_track_loc += 1
					# sample to use: unused for type 0xa0...?
					# either that or we don't include volume.  unclear what the magic byte here is for
					volume_maybe = (dtpkbytes[unk_track_loc:unk_track_loc+1])[0]
					unk_track_loc += 1
					trailing_bytes.append((dtpkbytes[unk_track_loc:unk_track_loc+1])[0])
					unk_track_loc += 1
					current_track.AddSample(entry_type, INVALID_SAMPLE_ID, volume_maybe, trailing_bytes)
				else:
					PrintAsFailure("ERROR! Unknown entry type %02x" % entry_type)

				allTracks.append(current_track)
				current_track.Print()
				unk_track_loc += 1
			
		allGroups.append(allTracks)
				
		# reset for pass 2
		sequencer_data_tables_offset = unk_track_loc
		# table 2 is padded to a 4 byte boundary, so adjust here
		spill = (sequencer_data_tables_offset % 4)
		if (spill != 0):
			sequencer_data_tables_offset += 4 - spill

	# OK, we're now past the sequencer data bank and into track to sample mapping
	# Starting at (sequencer_data_tables_offset) which now matches sample_playback_def_header_loc, there is 
	# a 0n80 / 0x50 sample playback definition header.  The only important here is at byte 0x10,
	# which is the number of sample playback definitions.
	# This is followed by 64 bytes per track
	track_data_table_loc = sequencer_data_tables_offset + 80

	if (sequencer_data_tables_offset != sample_playback_def_header_loc):
		PrintAsFailure('ERROR: the sequencer data table termination does not match the start of sample playback data.  This file should be looked at.')

	spdEntryCount_offset = sample_playback_def_header_loc + 0x10
	spdHeaderEntryCount = dtpkbytes[spdEntryCount_offset:spdEntryCount_offset + 0x2][0] + 1
	if show_detailed_data_playback:
		print ('\tSample playback entries (def at 0x%x): 0x%02x' % (spdEntryCount_offset, spdHeaderEntryCount))

	# Speculative format of the 64 byte track composition block 
	# Byte 00: track ID
	# Byte 01: Unknown, usually 80
	# Byte 02: sample bank ID used 
	# Byte 03: Unknown, only values 1 or 2 seem to be used
	# Byte 04: Unused?
	# Byte 05: ... offset?
	# Byte 06: Panning? Value 0x1f == left channel only.  Value 0x0f == right channel only.
		# Note that panning likely has 13 total steps
	# Byte 07: Volume?  Values 0x00-0x7f.  0x80 or higher uses default volume.
	# Byte 08: Volume?  Values 0x00-0xff.  Unsure of interaction with Byte 07
	# Byte 09: Rate/frequency?.  Often e8, f4
	# Byte 10: Unknown, usually 00.  Has no effect?
	# Byte 11: Unknown, usually 1f.  Has effect
	# Byte 12: Unknown, usually 00.  Has no effect?
	# Byte 13: Unknown, usually 1f.  Has no effect?
	# Byte 14: Unknown, usually 00.  Has no effect?
	# Byte 15: Unknown, usually 00.  Has effect
	# Byte 16: Unknown, usually 00.  Has no effect?
	# Byte 17: fd
	# Byte 18: 07
	# Byte 19: fd
	# Byte 20: 07
	# Byte 21: fd
	# Byte 22: 07
	# Byte 23: fd
	# Byte 24: 27
	# Byte 25: 00
	# Byte 26: 00
	# Byte 27: 00
	# Byte 28: 00
	# Byte 29: 00
	# Byte 30: 00
	# Byte 31: 7f has effect
	# Byte 32: 00
	# Byte 33: 00
	# Byte 34: 00
	# Byte 35: 00
	# Byte 36: 00
	# Byte 37: 00
	# Byte 38: 00
	# Byte 39: 00
	# Byte 40: 00
	# Byte 41: 00
	# Byte 42: 00
	# Byte 43: 00
	# Byte 44: 00
	# Byte 45: 00
	# Byte 46: 00
	# Byte 47: 80
	# Byte 48: 00
	# Byte 49: 00
	# Byte 50: 00
	# Byte 51: 00
	# Byte 52: 00
	# Byte 53: 00
	# Byte 54: 00
	# Byte 55: 00
	# Byte 56: 00
	# Byte 57: 00
	# Byte 58: 00
	# Byte 59: 00
	# Byte 60: 00
	# Byte 61: 00
	# Byte 62: 00
	# Byte 63: 00

	if show_detailed_data_playback:
		print ('\tSample playback data (64 bytes per sample):')

	allPlaybacks = []

	cur_track_loc = track_data_table_loc

	for curSamplePlaybackDef in range(nTotalSPDsUsed):
		playback_id = (dtpkbytes[cur_track_loc:cur_track_loc + 1])[0]
		validation_byte = (dtpkbytes[cur_track_loc + 4:cur_track_loc + 5])[0]
		if (validation_byte != 0x00):
			print ('Errata: track playback data entry %u is not valid: we\'re probably reading too far.' % curSamplePlaybackDef)
			continue
		# Note that there can be a conflict(?) between the Track Composition Data sample value and 
		# the Track Playback Data sample value.  For at least reDream, the non-zero value of the two is ignored.
		# In the instance where both are non-zero, the Track Composition Data value is used.
		playback_sample_used = (dtpkbytes[cur_track_loc + 2:cur_track_loc + 3])[0]
		playback_sample_rate = (dtpkbytes[cur_track_loc + 10:cur_track_loc + 11])[0]

		# Note that it would technically be more correct to use playback_id here, EXCEPT that value is often wrong
		cur_playback = DTPKPlaybackDef(curSamplePlaybackDef, cur_track_loc, playback_sample_rate, playback_sample_used)
		allPlaybacks.append(cur_playback)
		cur_playback.Print()

		if (playback_id != curSamplePlaybackDef):
			print ('\t\t\tErrata: Unexpected sample ID 0x%x: we\'ll go with the correct-er 0x%x' % (playback_id, curSamplePlaybackDef))
		cur_track_loc += 64

	countPCMSamples = 0
	countADPCMSamples = 0
	countADPCMExports = 0
	countWAVExports = 0

	dtpkOutBytes = array('B', dtpkbytes)
	injectionOffset = sample_offset + (16 * count_samples)

	if show_detailed_data_samples:
		print ('\tSample definition data:')
	for curSample in range(count_samples):
		# This parsing logic is based squarely upon https://github.com/Shenmue-Mods/DTPKUtil/blob/master/AM2-DTPK.txt
		# Any mistakes here are mine, any wisdom here is theirs
		cur_sample_offset_start = cur_sample_offset = sample_offset
		sample_flags = unpack('<I', dtpkbytes[cur_sample_offset:cur_sample_offset + 4])[0]
		cur_sample_offset  = cur_sample_offset + 4

		if (sample_flags & 0x01000000):
			countADPCMSamples += 1
		else:
			countPCMSamples += 1

		sample_loop_start = unpack('H', dtpkbytes[cur_sample_offset:cur_sample_offset + 2])[0]
		cur_sample_offset = cur_sample_offset + 2
		sample_loop_end = unpack('H', dtpkbytes[cur_sample_offset:cur_sample_offset + 2])[0]
		cur_sample_offset = cur_sample_offset + 2

		sample_rate = unpack('<I', dtpkbytes[cur_sample_offset:cur_sample_offset + 4])[0]
		cur_sample_offset = cur_sample_offset + 4
		sample_length = unpack('<I', dtpkbytes[cur_sample_offset:cur_sample_offset + 4])[0]

		actual_rate = 0
		playback_id = 0xFFFF
		track_id = 0xFFFF
		group_id = 0xFFFF
		for curPlayback in allPlaybacks:
			if (curPlayback.sample_used == curSample):
				actual_rate = curPlayback.rate
				playback_id = curPlayback.playback_id

				totalPlaybacksReferenced = 0

				for iGroupPos, curGroup in enumerate(allGroups):
					for iPos, curTrack in enumerate(curGroup):
						totalPlaybacksReferenced += curTrack.GetSampleCount()
						if (playback_id < totalPlaybacksReferenced):
							track_id = iPos
							group_id = iGroupPos
							break
					if (track_id != 0xFFFF):
						break
				break

		if (track_id == 0xFFFF):
			print ('Errata: playback ID for sample 0x%02x not found.  This file is using an unexpected structure.' % curSample)

		sampleData = DTPKSample(curSample, sample_offset, sample_flags, sample_loop_start, sample_loop_end, sample_rate, sample_length, group_id, track_id, playback_id, actual_rate)

		sampleData.Print()

		if (dump_adpcm_samples):
			countADPCMExports += sampleData.ExtractIfYADPCM(filename_base, dtpkbytes)
		if (dump_pcm_samples_to_wav):
			countWAVExports += sampleData.ExtractPCMToWAV(filename_base, dtpkbytes)

		# OK, we now have all the sample definition data: do we want to update it?
		fMustUpdateHeader = (len(changedSamples) != 0)
		for curUpdatedSample in changedSamples:
			if (curUpdatedSample.samplePosition == curSample):
				curSampleSize = len(curUpdatedSample.dataObject)
				print ('Changing sample %u to %s len 0x%x' % (curUpdatedSample.samplePosition, curUpdatedSample.sampleType, curSampleSize))

				print ('\tRebuilding sample definition entry at 0x%x' % cur_sample_offset_start)
				dtpkOutBytes[cur_sample_offset_start:cur_sample_offset_start + 16] = array('B', curUpdatedSample.GetHeader(injectionOffset))

				print ('\tSample injected inline at 0x%x to 0x%x, size change of %u' % (injectionOffset, injectionOffset + curSampleSize, curSampleSize - sampleData.length))
				dtpkOutBytes[injectionOffset:injectionOffset + curSampleSize] = array('B', curUpdatedSample.dataObject)
				injectionOffset += curSampleSize
				remnant = injectionOffset % 4
				if (remnant != 0):
					print ('\tAdding %u bytes padding (end of sample was 0x%x)' % (remnant, injectionOffset))
					dtpkOutBytes[injectionOffset:injectionOffset + remnant] = array('B', bytes(remnant))
					injectionOffset += remnant

				print ('\tUpdating sampling rate...')
				previousRate = INVALID_SAMPLE_ID
				warnedOnce = False
				for curPlayback in allPlaybacks:
					if (curPlayback.sample_used == curSample):
						thisOldRate = dtpkOutBytes[curPlayback.track_loc + 10:curPlayback.track_loc + 11][0]
						nUpdatedRate = TranslateDTPKRate(curUpdatedSample.rate)
						dtpkOutBytes[curPlayback.track_loc + 10:curPlayback.track_loc + 11] = array('B', nUpdatedRate.to_bytes(1))
						print ('\t\tUpdated sample table at 0x%x to DTPK rate 0x%x (frequency %u) from 0x%x' % (curPlayback.track_loc, nUpdatedRate, curUpdatedSample.rate, thisOldRate))
						if ((previousRate != INVALID_SAMPLE_ID) and (previousRate != thisOldRate) and (not warnedOnce)):
							PrintAsWarning('Warning: this sample is used at multiple rates.  You may want to further adjust the rate value(s) to account for that.')
							warnedOnce = True
						previousRate = thisOldRate

				fMustUpdateHeader = False

		sample_offset = cur_sample_offset + 4

		if (fMustUpdateHeader):
			dtpkOutBytes[injectionOffset:injectionOffset + sampleData.length] = dtpkbytes[sampleData.data_offset:sampleData.data_offset + sampleData.length]
			print ('\tAdjusting sample definition entry at 0x%x for sample location 0x%x length 0x%x' % (cur_sample_offset_start, injectionOffset, sampleData.length))
			dtpkOutBytes[cur_sample_offset_start:cur_sample_offset_start + 4] = array('B', sampleData.GetFlagValueForNewLocation(injectionOffset).to_bytes(4, byteorder='little'))
			injectionOffset += sampleData.length

	if ((mvc2_DTPK_size_limit < injectionOffset) and fIsProbablyMvC2CharacterFile):
		PrintAsFailure('ERROR: mvc2 character file size limit reached.')

	if ((not fIsShowingAnyInfo) and (len(outFile) == 0) and (not dump_adpcm_samples) and (not dump_pcm_samples_to_wav)):
		print ('\n"%s" is a valid DTPK file.  It contains %u command groups, %u tracks, and uses %u samples (%u PCM, %u ADPCM).' % (dtpkfile, ncmds, ntracks, count_samples, countPCMSamples, countADPCMSamples))
		print ('\tThis script can display detailed file information and/or extract out the samples.')
		print ('\tUse -options to see the current list of options.')
		print ('')

	if (show_detailed_file_info):
		print ('%u ADPCM samples found' % countADPCMSamples)
	if (countADPCMExports != 0):
		print ('Extracted out %u ADPCM samples' % countADPCMExports)
	elif (dump_adpcm_samples):
		print ('No ADPCM samples found.')

	if (show_detailed_file_info):
		print ('%u PCM samples found' % (countPCMSamples))
	if (countWAVExports != 0):
		print ('Extracted out %u PCM samples to WAV' % countWAVExports)
	elif (dump_pcm_samples_to_wav):
		print ('No PCM samples found.')

	if (len(outFile) != 0):
		# Update the filesize just in case
		if (injectionOffset != filesize_actual):
			# Enforce alignment
			remnant = injectionOffset % 4
			if (remnant != 0):
				print ('\tAdding %u bytes padding (size was 0x%x)' % (remnant, injectionOffset))
				dtpkOutBytes[injectionOffset:injectionOffset + remnant] = array('B', bytes(remnant))
				injectionOffset += remnant
			print ('\tUpdating header: new file size 0x%x' % injectionOffset)
			dtpkOutBytes[8:12] = array('B', injectionOffset.to_bytes(4, byteorder='little'))
			# truncate off the unused region, if any
			while (injectionOffset < len(dtpkOutBytes)):
				dtpkOutBytes.pop()

		if ((injectionOffset > mvc2_DTPK_size_limit) and fIsProbablyMvC2CharacterFile):
			PrintAsFailure('ERROR!  You have 2MB total for audio for Dreamcast.')
			print ('ERROR!  Keep file sizes for mvc2 characters to ~0x%x (%u) bytes to avoid overflow.' % (mvc2_DTPK_size_limit, mvc2_DTPK_size_limit))
			print ('ERROR!  If you cross that boundary, weird things will happen with sounds and/or you\'ll crash the Dreamcast.')
			print ('  If you\'re not worried about MvC2, you\'re probably fine.')
		elif (injectionOffset > (filesize_actual + 20000)):
			PrintAsWarning('Warning: These changes have grown the size of this audio file by %u bytes.  The Dreamcast has extremely limited memory.' % (injectionOffset - filesize_actual))
			print ('As such, downsampling injected files is usually smart if you want to avoid issues.')
		open(outFile, "wb").write(dtpkOutBytes)
		print ('\nFile \'%s\' created.' % outFile)

# =====================================================================================================

def PrintUsage():
	print ('\nAM2 DTPK info script v0.009')
	print ('Usage: python DTPKDump.py [options] <DTPK sound file>')
	print ('Options:')
	print ('          -showinfo         display detailed information about all DTPK contents')
	#print ('          -showtrackinfo    display detailed information about just the track data')
	#print ('          -showplaybackinfo display detailed information about just the sample playback data')
	#print ('          -showsampleinfo   display detailed information about just the samples')
	print ('          -extractall       extract out all samples')
	print ('          -wavconv          extract out all samples as PCM WAV files')
	#print ('          -dumpadpcm        extract out all ADPCM samples to raw Yamaha ADPCM files (usable in Awave Studio)')
	#print ('          -dumppcm          extract out all PCM samples to PCM WAV files (may not work for strange playback rates)')
	#print ('          -dumppcmall       extract out *all* samples to PCM WAV files (may not work for strange playback rates)')
	print ('                            (This options uses Sappharad\'s ADPCM to PCM conversion code.)')
	print ('          -spliceadpcm (NUM) (FREQUENCY) (FILENAME)  replace sample NUM in the DTPK with the specified ADPCM file')
	print ('                            Requires -out')
	print ('          -splicepcm (NUM) (FILENAME)  replace sample NUM in the DTPK with the specified PCM WAV file')
	print ('                            WAV files must be PCM 8bit or 16bit at either 11025 or 22050 sample rate')
	print ('                            Requires -out')
	print ('          -out (FILENAME)   save the updated DTPK file as FILENAME')
	print ('                            Requires one of the splicing options.')
	

# =====================================================================================================
if __name__ == '__main__':
	argv = sys.argv
	argc = len(argv)
	os.system('color')

	fFailedArgProcessing = False
	dtpkFilename = ""
	outFile = ""

	changedSamples = []

	if (argc > 1):
		iCurrArg = 1
		while iCurrArg <= len(argv[1:]):
			if (argv[iCurrArg][0] == '-'):
				if (argv[iCurrArg] == "-showinfo"):
					show_detailed_file_info = True
					show_detailed_data_tracks = True
					show_detailed_data_playback = True
					show_detailed_data_samples = True
					show_detailed_data_samples_full = True
				elif (argv[iCurrArg] == "-showtrackinfo"):
					show_detailed_data_tracks = True
				elif (argv[iCurrArg] == "-showplaybackinfo"):
					show_detailed_data_playback = True
				elif (argv[iCurrArg] == "-showsampleinfo"):
					show_detailed_data_samples = True
					show_detailed_data_samples_full = True
				elif ((argv[iCurrArg] == "-dump") or (argv[iCurrArg] == "-dumpall") or (argv[iCurrArg] == "-extractall")):
					dump_adpcm_samples = True
					dump_pcm_samples_to_wav = True
				elif (argv[iCurrArg] == "-dumpadpcm"):
					dump_adpcm_samples = True
				elif (argv[iCurrArg] == "-dumppcm"):
					dump_pcm_samples_to_wav = True
				elif ((argv[iCurrArg] == "-dumppcmall") or (argv[iCurrArg] == "-wavconv")):
					dump_pcm_samples_to_wav = True
					dump_adpcm_samples_to_wav = True
				elif (argv[iCurrArg] == "-spliceadpcm"):
					iCurrArg += 1
					try:
						samplePosition = int(argv[iCurrArg])
					except:
						try:
							samplePosition = int(argv[iCurrArg], 16)
						except:
							PrintAsFailure('ERROR: check usage.  You need to tell us what sample slot to inject this sample into.\n')
							fFailedArgProcessing = True
							break
					iCurrArg += 1
					try:
						sampleRate = int(argv[iCurrArg])
					except:
						try:
							sampleRate = int(argv[iCurrArg], 16)
						except:
							PrintAsFailure('ERROR: check usage.  You need to tell us what sample slot to inject this sample into.\n')
							fFailedArgProcessing = True
							break
					iCurrArg += 1
					try:
						adpcmFile = argv[iCurrArg]
					except:
						PrintAsFailure('ERROR: check usage.  You need to tell us what ADPCM file to splice in.')
						fFailedArgProcessing = True
						break
					
					if (not ImportAndAddADPCMFile(adpcmFile, changedSamples, samplePosition, sampleRate)):
						PrintAsFailure('ERROR: file \'%s\' could not be imported.' % adpcmFile)
						fFailedArgProcessing = True
						break
				elif (argv[iCurrArg] == "-splicepcm"):
					iCurrArg += 1
					try:
						samplePosition = int(argv[iCurrArg])
					except:
						try:
							samplePosition = int(argv[iCurrArg], 16)
						except:
							PrintAsFailure('ERROR: check usage.  You need to tell us what sample slot to inject this sample into.\n')
							fFailedArgProcessing = True
							break
					iCurrArg += 1
					try:
						wavFile = argv[iCurrArg]
					except:
						PrintAsFailure('ERROR: check usage.  You need to tell us what WAV file to splice in.')
						fFailedArgProcessing = True
						break
					
					if (not ImportAndAddWAVFile(wavFile, changedSamples, samplePosition)):
						PrintAsFailure('ERROR: file \'%s\' could not be imported.' % wavFile)
						fFailedArgProcessing = True
						break
				elif ((argv[iCurrArg] == "-o") or (argv[iCurrArg] == "-out")):
					iCurrArg += 1
					outFile = argv[iCurrArg]
				elif ((argv[iCurrArg] == "-options") or (argv[iCurrArg] == "-?")):
					break
				else:
					PrintAsFailure('Error: unsupported option %s' % argv[iCurrArg])
					fFailedArgProcessing = True
					break
			else:
				dtpkFilename = argv[iCurrArg]
			iCurrArg += 1

	if ((len(changedSamples) != 0) and (len(outFile) == 0)):
		PrintAsFailure('ERROR: you must specify an output filename if trying to splice in a sample.\n')
		dtpkFilename = ""
	if ((len(changedSamples) == 0) and (len(outFile) != 0)):
		PrintAsFailure('ERROR: output file specified with no samples to splice in.\n')
		dtpkFilename = ""

	if ((len(dtpkFilename) != 0) and (not fFailedArgProcessing)):
		dtpkparse(dtpkFilename, outFile, changedSamples)
	else:
		PrintUsage()
		sys.exit(0)

# =====================================================================================================

# =====================================================================================================
# 24-05-18 (0.009) - Mapped out more of the DTPK structure
# 24-04-25 (0.008) - Improve handling of sequencer groups
# 24-04-20 (0.007) - Allow exporting YADPCM chunks as RIFF/WAV PCM
# 24-04-16 (0.006) - Significantly improve rate handling
# 24-04-13 (0.005) - Efficiency: always inline sample replacement
# 24-04-10 (0.004) - Support splicing in ADPCM chunks or PCM WAV files
# 24-04-03 (0.003) - Support slightly limited PCM export to WAV
# 24-03-27 (0.002) - Smooth out T->SPD->S logic, filenames, etc.
# 24-03-21 (0.001) - Add optional sample extraction
# 24-03-01 (0.000) - Initial version based upon KingShriek and Shenmue's knowledge.
# =====================================================================================================
